import ssl
from typing import Literal

SSL_CONTEXT = ssl._create_unverified_context()
DEFAULT_REST_API_HOST = "api-global-prod.aircloudhome.com"
DEFAULT_STOMP_WEBSOCKET_HOST = "notification-global-prod.aircloudhome.com"
DEFAULT_WAIT_DONE = 10

CommandState = Literal["SENDING", "INCOMPLETE", "DONE"]

FanSwing = Literal["OFF", "VERTICAL", "HORIZONTAL", "BOTH", "AUTO"]
FanSpeed = Literal["LV1", "LV2", "LV3", "LV4", "LV5", "AUTO"]
Power = Literal["ON", "OFF"]

# Don't know what is DRY_COOL but it exist in Android client source code
OperatingMode = Literal["AUTO", "COOLING", "DE_HUMIDIFY", "DRY_COOL", "FAN", "HEATING"]
TemperatureUnit = Literal["CELSIUS", "FAHRENHEIT"]

ScheduleType = Literal[
    "SCHEDULE_DISABLED",
    "OFF_TIMER_ENABLED",
    "ON_TIMER_ENABLED",
    "ON_OFF_TIMER_ENABLED",
    "WEEKLY_TIMER_ENABLED",
    "HOLIDAY_MODE_ENABLED",
]



from typing import Optional, Tuple

from aircloudy.utils import to_int

from .frames_models import StompFrame


class ConnectedFrame(StompFrame):
    version: str
    heart_beat: Optional[Tuple[int, int]]
    session: Optional[str]
    server: Optional[str]

    def __init__(self, frame: StompFrame) -> None:
        StompFrame.__init__(
            self,
            frame.message,
            frame.headers,
            frame.body,
        )
        self.version = frame.headers["version"]

        heart_beat_as_string: str = frame.headers["heart-beat"]
        if heart_beat_as_string is not None:
            heart_beat_min, heart_beat_max = heart_beat_as_string.split(",", 2)
            self.heart_beat = (int(heart_beat_min), int(heart_beat_max))

        self.session = frame.headers.get("session")
        self.server = frame.headers.get("server")


class MessageFrame(StompFrame):
    destination: str
    message_id: str
    subscription: str
    ack: Optional[str]
    content_type: Optional[str]
    content_length: Optional[int]

    def __init__(self, frame: StompFrame) -> None:
        StompFrame.__init__(
            self,
            frame.message,
            frame.headers,
            frame.body,
        )
        self.destination = frame.headers["destination"]
        self.message_id = frame.headers["message-id"]
        self.subscription = frame.headers["subscription"]
        self.ack = frame.headers.get("ack")
        self.content_type = frame.headers.get("content-type")
        self.content_length = to_int(frame.headers.get("content-length"))


class ReceiptFrame(StompFrame):
    receipt_id: str

    def __init__(self, frame: StompFrame) -> None:
        StompFrame.__init__(
            self,
            frame.message,
            frame.headers,
            frame.body,
        )
        self.receipt_id = frame.headers["receipt-id"]


class ErrorFrame(StompFrame):
    error_message: Optional[str]
    content_type: Optional[str]
    content_length: Optional[int]

    def __init__(self, frame: StompFrame) -> None:
        StompFrame.__init__(
            self,
            frame.message,
            frame.headers,
            frame.body,
        )
        self.error_message = frame.headers.get("message")
        self.content_type = frame.headers.get("content-type")
        self.content_length = to_int(frame.headers.get("content-length"))



import json

from aircloudy.contants import DEFAULT_REST_API_HOST

from ..errors import AuthenticationFailedException
from .http_client import perform_request
from .iam_models import AuthenticationSuccess, UserProfile


async def perform_login(
    email: str, password: str, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> AuthenticationSuccess:
    response = await perform_request(
        "POST",
        "/iam/auth/sign-in",
        {
            "email": email,
            "password": password,
        },
        do_not_raise_exception_on=(200, 401),
        host=host,
        port=port,
    )

    if response.status == 401:
        raise AuthenticationFailedException(f"Autentication Failed: {response.body_as_json.get('errorState')}")

    return json.loads(response.body, object_hook=AuthenticationSuccess)


async def fetch_profile(token: str, host: str = DEFAULT_REST_API_HOST, port: int = 443) -> UserProfile:
    response = await perform_request("GET", "/iam/user/v2/who-am-i", token=token, host=host, port=port)

    return json.loads(response.body, object_hook=UserProfile)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from .contants import FanSpeed, FanSwing, OperatingMode, Power


@dataclass
class InteriorUnit:
    id: int
    name: str
    power: Power
    mode: OperatingMode
    requested_temperature: float
    humidity: int  # don't know the effect/working, is it current or requested, ?
    fan_speed: FanSpeed
    fan_swing: FanSwing
    room_temperature: float
    updated_at: int
    online: bool
    online_updated_at: int

    def copy(
        self,
        name: Optional[str] = None,
        power: Optional[Power] = None,
        mode: Optional[OperatingMode] = None,
        requested_temperature: Optional[float] = None,
        humidity: Optional[int] = None,
        fan_speed: Optional[FanSpeed] = None,
        fan_swing: Optional[FanSwing] = None,
        room_temperature: Optional[float] = None,
        updated_at: Optional[int] = None,
        online: Optional[bool] = None,
        online_updated_at: Optional[int] = None,
    ) -> InteriorUnit:
        new_name = self.name
        new_power = self.power
        new_mode = self.mode
        new_requested_temperature = self.requested_temperature
        new_humidity = self.humidity
        new_fan_speed = self.fan_speed
        new_fan_swing = self.fan_swing
        new_room_temperature = self.room_temperature
        new_updated_at = self.updated_at
        new_online = self.online
        new_online_updated_at = self.online_updated_at

        if name is not None:
            new_name = name
        if power is not None:
            new_power = power
        if mode is not None:
            new_mode = mode
        if requested_temperature is not None:
            new_requested_temperature = requested_temperature
        if humidity is not None:
            new_humidity = humidity
        if fan_speed is not None:
            new_fan_speed = fan_speed
        if fan_swing is not None:
            new_fan_swing = fan_swing
        if room_temperature is not None:
            new_room_temperature = room_temperature
        if updated_at is not None:
            new_updated_at = updated_at
        if online is not None:
            new_online = online
        if online_updated_at is not None:
            new_online_updated_at = online_updated_at

        return InteriorUnit(
            self.id,
            new_name,
            new_power,
            new_mode,
            new_requested_temperature,
            new_humidity,
            new_fan_speed,
            new_fan_swing,
            new_room_temperature,
            new_updated_at,
            new_online,
            new_online_updated_at,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InteriorUnit):
            return False

        return (
            self.id == other.id
            and self.name == other.name
            and self.power == other.power
            and self.mode == other.mode
            and self.requested_temperature == other.requested_temperature
            and self.humidity == other.humidity
            and self.fan_speed == other.fan_speed
            and self.fan_swing == other.fan_swing
            and self.room_temperature == other.room_temperature
            and self.updated_at == other.updated_at
            and self.online == other.online
            and self.online_updated_at == other.online_updated_at
        )



from dataclasses import dataclass
from typing import List, Optional

from aircloudy.contants import CommandState, FanSpeed, FanSwing, OperatingMode, Power, ScheduleType
from aircloudy.interior_unit_models import InteriorUnit


class CommandResponse:
    commandId: str
    thingId: str

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


class CommandStatus:
    commandId: str
    status: CommandState

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


class PowerResult:
    racId: int
    success: bool
    errorMessage: Optional[str]
    errorCode: int
    commandResponse: CommandResponse

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


class PowerAllResponse:
    allSucceeded: bool
    resultSet: List[PowerResult]

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


@dataclass
class _GeneralControlCommand:
    id: int
    power: Power
    mode: OperatingMode
    iduTemperature: float
    humidity: int
    fanSpeed: FanSpeed
    fanSwing: FanSwing

    def __init__(self, interior_unit: InteriorUnit) -> None:
        self.id = interior_unit.id
        self.power = interior_unit.power
        self.mode = interior_unit.mode
        self.iduTemperature = interior_unit.requested_temperature
        # Should be something like `self.humidity = interior_unit.humidity`
        # but some api return me humidity=126 and this is not a legal value for control
        self.humidity = 50
        self.fanSpeed = interior_unit.fan_speed
        self.fanSwing = interior_unit.fan_swing


@dataclass
class _InteriorUnitRest:
    userId: str
    serialNumber: str
    model: str
    id: int
    vendorThingId: str
    name: str
    roomTemperature: float
    mode: OperatingMode
    iduTemperature: float
    humidity: int
    power: Power
    relativeTemperature: float
    fanSpeed: FanSpeed
    fanSwing: FanSwing
    updatedAt: int
    lastOnlineUpdatedAt: int
    racTypeId: int
    iduFrostWash: bool
    specialOperation: bool
    criticalError: bool
    zoneId: str
    scheduleType: ScheduleType
    online: bool

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)

    def to_internal_representation(self) -> InteriorUnit:
        return InteriorUnit(
            self.id,
            self.name,
            self.power,
            self.mode,
            self.iduTemperature,
            self.humidity,
            self.fanSpeed,
            self.fanSwing,
            self.roomTemperature,
            self.updatedAt,
            self.online,
            self.lastOnlineUpdatedAt,
        )



from uuid import UUID

from .stomp import StompFrame


class ConnectFrame(StompFrame):
    def __init__(self, token: str) -> None:
        StompFrame.__init__(
            self,
            "CONNECT",
            {
                "accept-version": "1.1,1.2",
                "heart-beat": "10000,10000",
                "Authorization": f"Bearer {token}",
            },
        )


class RefreshAllInteriorUnitFrame(StompFrame):
    def __init__(self, token: str, user_id: int, family_id: int) -> None:
        StompFrame.__init__(
            self,
            "MESSAGE",
            {
                "Authorization": f"Bearer {token}",
                "destination": f"/app/racs/{user_id}/{family_id}",
            },
            {
                "racId": 0,
                "requestType": "REFRESH_ALL",
            },
        )


class RefreshInteriorUnitFrame(StompFrame):
    def __init__(self, token: str, user_id: int, family_id: int, rac_id: int) -> None:
        StompFrame.__init__(
            self,
            "MESSAGE",
            {
                "Authorization": f"Bearer {token}",
                "destination": f"/app/racs/{user_id}/{family_id}",
            },
            {
                "racId": rac_id,
                "requestType": "REFRESH_INDIVIDUAL",
            },
        )


class SubscribeFrame(StompFrame):
    def __init__(self, uuid: UUID, user_id: int, family_id: int) -> None:
        StompFrame.__init__(
            self,
            "SUBSCRIBE",
            {
                "id": str(uuid),
                "destination": f"/notification/{user_id}/{family_id}",
                "ack": "auto",
            },
        )


class UnsubscribeFrame(StompFrame):
    def __init__(self, uuid: UUID) -> None:
        StompFrame.__init__(
            self,
            "UNSUBSCRIBE",
            {
                "id": str(uuid),
            },
        )



import copy
import logging
import uuid
from dataclasses import dataclass
from types import TracebackType
from typing import Callable, List, Optional, Self, Tuple, Type

from . import api, notifications
from .api import CommandResponse
from .contants import (
    DEFAULT_REST_API_HOST,
    DEFAULT_STOMP_WEBSOCKET_HOST,
    DEFAULT_WAIT_DONE,
    FanSpeed,
    FanSwing,
    OperatingMode,
    Power,
    TemperatureUnit,
)
from .errors import CommandFailedException, IllegalStateException, InteriorUnitNotFoundException
from .interior_unit_models import InteriorUnit

logger = logging.getLogger(__name__)


@dataclass
class ConnectionInfo:
    authentication_result: api.AuthenticationSuccess
    user_profile: api.UserProfile
    notification_socket: notifications.NotificationsWebsocket
    notification_subscription_id: uuid.UUID


class HitachiAirCloud:
    _email: str
    _password: str
    _default_wait_done: int
    _api_host: str

    _command_manager: api.CommandManager
    _connection_info: Optional[ConnectionInfo]
    _interior_units: dict[int, InteriorUnit]

    on_change: Optional[Callable[[dict[int, Tuple[Optional[InteriorUnit], Optional[InteriorUnit]]]], None]]

    def __init__(
        self,
        email: str,
        password: str,
        default_wait_done: int = DEFAULT_WAIT_DONE,
        api_host: str = DEFAULT_REST_API_HOST,
        api_port: int = 443,
        notification_host: str = DEFAULT_STOMP_WEBSOCKET_HOST,
    ) -> None:
        self._email = email
        self._password = password
        self._default_wait_done = default_wait_done
        self._api_host = api_host
        self.notification_host = notification_host

        self._command_manager = api.CommandManager(self._get_auth_token_or_fail, host=api_host, port=api_port)
        self._connection_info = None
        self._interior_units = {}

        self.on_change = None

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        await self.close()
        return False

    @property
    def is_open(self) -> bool:
        return self._connection_info is not None

    @property
    def interior_units(self) -> List[InteriorUnit]:
        return [copy.deepcopy(iu) for iu in self._interior_units.values()]

    @property
    def temperature_unit(self) -> TemperatureUnit:
        if self._connection_info is None:
            raise Exception("AirCloud is not connected")

        if self._connection_info.user_profile.settings.temperatureUnit == "degC":
            return "CELSIUS"

        return "FAHRENHEIT"

    def interior_unit(self, rac_id: int) -> Optional[InteriorUnit]:
        return self._interior_units.get(rac_id)

    def _get_auth_token_or_fail(self) -> str:
        if self._connection_info is None:
            raise Exception("AirCloud is not connected")
        return self._connection_info.authentication_result.token

    async def _wait_done(self, override_default_wait_done: Optional[int], command: CommandResponse) -> None:
        wait_done = self._default_wait_done if override_default_wait_done is None else override_default_wait_done
        if wait_done > 0:
            final_state = await self._command_manager.wait_ack(command, wait_done)
            if final_state != "DONE":
                raise CommandFailedException(
                    f"Command state is still not done (current: {final_state}) after delay {wait_done}"
                )

    async def connect(self) -> None:
        if self.is_open:
            raise IllegalStateException("AirCloud already connected")

        authentication_result = await api.perform_login(self._email, self._password, self._api_host)
        user_profile = await api.fetch_profile(authentication_result.token, self._api_host)
        self._interior_units = {
            iu.id: iu
            for iu in await api.get_interior_units(authentication_result.token, user_profile.familyId, self._api_host)
        }

        notification_socket = notifications.NotificationsWebsocket(
            self.notification_host,
            lambda: authentication_result.token,
            user_profile.id,
            user_profile.familyId,
            lambda ius: self._update_interior_units(ius),
        )
        await notification_socket.connect()
        notification_subscription_id = await notification_socket.subscribe()

        self._connection_info = ConnectionInfo(
            authentication_result,
            user_profile,
            notification_socket,
            notification_subscription_id,
        )

        notification_socket.token_supplier = self._get_auth_token_or_fail
        logger.info("Connected")

    async def close(self) -> None:
        try:
            if self._connection_info is not None:
                await self._connection_info.notification_socket.close()
        finally:
            self._connection_info = None
            self._interior_units = {}
        logging.info("Closed")

    def _update_interior_units(self, interior_units: List[InteriorUnit]) -> None:
        logger.debug("Received interior units update: %s", interior_units)
        changes: dict[int, Tuple[Optional[InteriorUnit], Optional[InteriorUnit]]] = {}
        for iu in interior_units:
            old_iu = self._interior_units[iu.id]
            if old_iu != iu:
                changes[iu.id] = (old_iu, iu)
            self._interior_units[iu.id] = iu

        if self.on_change is not None:
            self.on_change(changes)

    async def set(self, interior_unit: InteriorUnit, override_default_wait_done: Optional[int] = None) -> None:
        """
        Send command to set interior_unit state

        :raises:
            IllegalStateException: If instance is not connected
            CommandFailedException: If state isn't done after wait delay
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        command: CommandResponse = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            interior_unit,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)

    async def set_power(
        self, interior_unit: InteriorUnit, power: Power, override_default_wait_done: Optional[int] = None
    ) -> None:
        """
        Send command to set interior_unit on/off

        :raises:
            IllegalStateException: If instance is not connected
            InteriorUnitNotFoundException: If given interior_unit id doesn't exists in current know states
            CommandFailedException: If state isn't done after wait delay
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        up_to_date_interior_unit = self.interior_unit(interior_unit.id)
        if up_to_date_interior_unit is None:
            raise InteriorUnitNotFoundException(f"Interior unit {interior_unit.id} not found")

        command = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            up_to_date_interior_unit,
            power=power,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)

    async def set_operating_mode(
        self, interior_unit: InteriorUnit, mode: OperatingMode, override_default_wait_done: Optional[int] = None
    ) -> None:
        """
        Send command to set interior_unit operating mode

        :raises:
            IllegalStateException: If instance is not connected
            InteriorUnitNotFoundException: If given interior_unit id doesn't exists in current know states
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        up_to_date_interior_unit = self.interior_unit(interior_unit.id)
        if up_to_date_interior_unit is None:
            raise InteriorUnitNotFoundException(f"Interior unit {interior_unit.id} not found")

        command = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            up_to_date_interior_unit,
            mode=mode,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)

    async def set_requested_temperature(
        self, interior_unit: InteriorUnit, temperature: float, override_default_wait_done: Optional[int] = None
    ) -> None:
        """
        Send command to set interior_unit requested temperature

        :raises:
            IllegalStateException: If instance is not connected
            InteriorUnitNotFoundException: If given interior_unit id doesn't exists in current know states
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        up_to_date_interior_unit = self.interior_unit(interior_unit.id)
        if up_to_date_interior_unit is None:
            raise InteriorUnitNotFoundException(f"Interior unit {interior_unit.id} not found")

        command = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            up_to_date_interior_unit,
            requested_temperature=temperature,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)

    async def set_humidity(
        self, interior_unit: InteriorUnit, humidity: int, override_default_wait_done: Optional[int] = None
    ) -> None:
        """
        Send command to set interior_unit humidity

        :raises:
            IllegalStateException: If instance is not connected
            InteriorUnitNotFoundException: If given interior_unit id doesn't exists in current know states
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        up_to_date_interior_unit = self.interior_unit(interior_unit.id)
        if up_to_date_interior_unit is None:
            raise InteriorUnitNotFoundException(f"Interior unit {interior_unit.id} not found")

        command = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            up_to_date_interior_unit,
            humidity=humidity,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)

    async def set_fan_speed(
        self, interior_unit: InteriorUnit, speed: FanSpeed, override_default_wait_done: Optional[int] = None
    ) -> None:
        """
        Send command to set interior_unit fan speed

        :raises:
            IllegalStateException: If instance is not connected
            InteriorUnitNotFoundException: If given interior_unit id doesn't exists in current know states
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        up_to_date_interior_unit = self.interior_unit(interior_unit.id)
        if up_to_date_interior_unit is None:
            raise InteriorUnitNotFoundException(f"Interior unit {interior_unit.id} not found")

        command = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            up_to_date_interior_unit,
            fan_speed=speed,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)

    async def set_fan_swing(
        self, interior_unit: InteriorUnit, swing: FanSwing, override_default_wait_done: Optional[int] = None
    ) -> None:
        """
        Send command to set interior_unit fan swing

        :raises:
            IllegalStateException: If instance is not connected
            InteriorUnitNotFoundException: If given interior_unit id doesn't exists in current know states
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        up_to_date_interior_unit = self.interior_unit(interior_unit.id)
        if up_to_date_interior_unit is None:
            raise InteriorUnitNotFoundException(f"Interior unit {interior_unit.id} not found")

        command = await api.configure_interior_unit(
            self._connection_info.authentication_result.token,
            self._connection_info.user_profile.familyId,
            up_to_date_interior_unit,
            fan_swing=swing,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self._connection_info.notification_socket.refresh(interior_unit.id)



from .command_manager import CommandManager
from .iam import fetch_profile, perform_login
from .iam_models import AuthenticationSuccess, UserProfile
from .rac import (
    configure_interior_unit,
    get_interior_units,
    request_refresh_interior_unit_state,
    set_power,
    set_power_all,
)
from .rac_models import CommandResponse



import asyncio
import logging
import uuid
from asyncio import Task
from types import TracebackType
from typing import Callable, List, Optional, Self, Type

import websockets

from aircloudy.contants import SSL_CONTEXT
from aircloudy.errors import IllegalStateException
from aircloudy.utils import current_task_is_running

from ..interior_unit_models import InteriorUnit
from . import hitachi_frame_models, stomp
from .interior_unit_models import InteriorUnitNotification

logger = logging.getLogger(__name__)


class NotificationsWebsocket:
    notification_host: str
    token_supplier: Callable[[], str]
    user_id: int
    family_id: int
    state_callback: Callable[[List[InteriorUnit]], None]
    on_connection_closed_by_server: Optional[Callable[[websockets.ConnectionClosed], None]]

    notification_socket: Optional[websockets.WebSocketClientProtocol] = None
    task_send_client_heartbeat: Optional[Task] = None
    task_handle_incoming_frame: Optional[Task] = None

    def __init__(
        self,
        notification_host: str,
        token_supplier: Callable[[], str],
        user_id: int,
        family_id: int,
        state_callback: Callable[[List[InteriorUnit]], None],
        on_connection_closed_by_server: Optional[Callable[[websockets.ConnectionClosed], None]] = None,
    ) -> None:
        self.notification_host = notification_host
        self.token_supplier = token_supplier
        self.user_id = user_id
        self.family_id = family_id
        self.state_callback = state_callback
        self.on_connection_closed_by_server = on_connection_closed_by_server
        self.notification_subscription_id = uuid.uuid4()

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        await self.close()
        return False

    @property
    def is_open(self) -> bool:
        return self.notification_socket is not None

    async def connect(self) -> None:
        if self.is_open:
            raise Exception(__name__ + " already connected")

        websocket_url = f"wss://{self.notification_host}/rac-notifications/websocket"
        logger.info("Open websocket to %s", websocket_url)
        self.notification_socket = await websockets.connect(websocket_url, ssl=SSL_CONTEXT)
        logger.debug("Send CONNECT stomp frame")
        await self.notification_socket.send(hitachi_frame_models.ConnectFrame(self.token_supplier()).get_frame())

        logger.debug("Wait first server frame (expect CONNECTED frame)")
        frame_data = await self.notification_socket.recv()
        if isinstance(frame_data, bytes):
            raise Exception("Binary frame was unexpected")
        first_server_frame = stomp.parse_server_frame(frame_data)
        if not isinstance(first_server_frame, stomp.ConnectedFrame):
            raise Exception(f"Excepted stomp.ConnectedFrame but got {first_server_frame}")

        logger.debug("Start send client heartbeat task")
        self.task_send_client_heartbeat = asyncio.create_task(self._send_client_heartbeat_loop())
        logger.debug("Start handle incoming frame task")
        self.task_handle_incoming_frame = asyncio.create_task(self._handle_incoming_frame_loop())

    async def subscribe(self) -> uuid.UUID:
        if self.notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        subscription_id = uuid.uuid4()
        logger.info(
            "Create subscription %s to notifications from user_id=%s, family_id=%s",
            subscription_id,
            self.user_id,
            self.family_id,
        )
        payload = hitachi_frame_models.SubscribeFrame(subscription_id, self.user_id, self.family_id)
        await self.notification_socket.send(payload.get_frame())
        return subscription_id

    async def unsubscribe(self, subscription_id: uuid.UUID) -> None:
        if self.notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        logger.info("Remove subscription %s", subscription_id)
        payload = hitachi_frame_models.UnsubscribeFrame(subscription_id)
        await self.notification_socket.send(payload.get_frame())

    async def refresh_all(self) -> None:
        if self.notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        logger.info("Request refresh all")
        payload = hitachi_frame_models.RefreshAllInteriorUnitFrame(self.token_supplier(), self.user_id, self.family_id)
        await self.notification_socket.send(payload.get_frame())

    async def refresh(self, rac_id: int) -> None:
        if self.notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        logger.info("Request refresh rac_id=%d", rac_id)
        payload = hitachi_frame_models.RefreshInteriorUnitFrame(
            self.token_supplier(), self.user_id, self.family_id, rac_id
        )
        await self.notification_socket.send(payload.get_frame())

    async def _send_client_heartbeat_loop(self) -> None:
        while current_task_is_running() and self.notification_socket is not None:
            try:
                logger.debug("Send heartbeat")
                await self.notification_socket.send("\r\n")
                await asyncio.sleep(10)
            except Exception as e:
                logger.error("Unexpected error while sending client heartbeat : %s", e)
                raise e

    async def _handle_incoming_frame_loop(self) -> None:
        while current_task_is_running() and self.notification_socket is not None:
            try:
                logger.debug("Waiting for message")
                data = await self.notification_socket.recv()
                if isinstance(data, bytes):
                    raise Exception("Binary frame was unexpected")

                frame = stomp.parse_server_frame(data)
                logger.debug("Received frame %s", frame)
                match frame:
                    case None:
                        logger.debug("Frame was server-heartbeat")
                    case stomp.ConnectedFrame():
                        raise Exception("ConnectedFrame should have been received at initialization")
                    case stomp.MessageFrame():
                        if frame.body is None:
                            raise Exception("Unexpected message without body")

                        notification_type = frame.body.get("notificationType")
                        if notification_type is None:
                            raise Exception("Unexpected message without notificationType")

                        if notification_type in ("ON_CONNECT", "BUCKET_UPDATE"):
                            interior_units = [
                                InteriorUnitNotification(d).to_internal_representation() for d in frame.body["data"]
                            ]
                            self.state_callback(interior_units)
                        else:
                            raise Exception("Unexpected message notification_type", notification_type)

                    case _:
                        logger.warning("Unexpected frame type : %s", frame.message)
            except websockets.ConnectionClosed as e:
                logger.debug("Connection closed (%s)", e)
                if self.on_connection_closed_by_server is not None:
                    self.on_connection_closed_by_server(e)

                # Not sure how to handle the situation knowing
                # that self.close will want the current task to be canceled
                # And the cancellation wait for this to finish
                asyncio.create_task(self.close())
                return
            except Exception as e:
                logger.error("Unexpected error while handling incoming message : %s", e)
                raise e

    async def close(self) -> None:
        try:
            if self.task_send_client_heartbeat is not None:
                self.task_send_client_heartbeat.cancel()
                try:
                    await self.task_send_client_heartbeat
                except asyncio.exceptions.CancelledError as e:
                    logger.debug("Canceled `send client heartbeat` task : %s", e)

            if self.task_handle_incoming_frame is not None:
                self.task_handle_incoming_frame.cancel()
                try:
                    await self.task_handle_incoming_frame
                except asyncio.exceptions.CancelledError as e:
                    logger.debug("Canceled `handle incoming frame` task : %s", e)

            if self.notification_socket is not None:
                await self.notification_socket.close()
                logger.info("Websocket closed")
        finally:
            self.notification_socket = None
            self.task_handle_incoming_frame = None
            self.task_send_client_heartbeat = None



from dataclasses import dataclass
from typing import List, Literal


@dataclass
class UserProfile:
    @dataclass
    class Settings:
        outOfHomeAddress: None
        sensitiveToCold: bool
        temperatureUnit: Literal["degC"]
        outOfHomeLongitude: float
        homeOnWeekdays: bool
        language: Literal["en"]
        outOfHomeRadius: float
        homeOnWeekends: bool
        outOfHomeRemainderEnabled: bool
        outOfHomeLatitude: float

        def __init__(self, data: dict) -> None:
            self.__dict__.update(data)

    @dataclass
    class Address:
        zipCode: str
        city: str
        street: str
        countryCode: str
        state: str
        addressLine: str

        def __init__(self, data: dict) -> None:
            self.__dict__.update(data)

    @dataclass
    class Role:
        level: int
        name: Literal["OWNER"]
        id: int

        def __init__(self, data: dict) -> None:
            self.__dict__.update(data)

    familyId: int
    firstName: str
    lastName: str
    settings: Settings
    address: Address
    phoneNumber: None
    pictureData: None
    familyName: str
    roles: List[Role]
    middleName: None
    id: int
    email: str

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


@dataclass
class AuthenticationSuccess:
    token: str
    refreshToken: str
    newUser: bool
    errorState: Literal["NONE"]
    access_token_expires_in: int
    refresh_token_expires_in: int

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)



import json
from typing import Optional


class IllegalStateException(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)


class TooManyRequestsException(Exception):
    error_type: str
    error_desc: str
    error_stack_trace: Optional[str]
    error_code: Optional[str]

    def __init__(self, body: str) -> None:
        data = json.loads(body)
        self.error_type = data["type"]
        self.error_desc = data["desc"]
        self.error_stack_trace = data.get("strackTrace")
        self.error_code = data.get("code")
        Exception.__init__(self, f"Too many requests: {self.error_desc}")


class InteriorUnitNotFoundException(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)


class CommandFailedException(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)


class AuthenticationFailedException(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)


class ConnectionFailed(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)



import logging
from typing import Literal, Optional, Tuple

import aiohttp
from aiohttp import TCPConnector

from aircloudy.contants import DEFAULT_REST_API_HOST, SSL_CONTEXT
from aircloudy.errors import ConnectionFailed

from .http_client_models import HttpResponse

logger = logging.getLogger(__name__)


def create_headers(host: str, token: Optional[str] = None) -> dict[str, str]:
    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json; charset=UTF-8",
        "Host": host,
        "User-Agent": "okhttp/4.2.2",
    }
    if token:
        headers["Authorization"] = f"Bearer {token}"
    return headers


async def perform_request(
    method: Literal["GET", "POST", "PUT"],
    url: str,
    body: Optional[object] = None,
    do_not_raise_exception_on: Tuple[int, ...] = (200,),
    token: Optional[str] = None,
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> HttpResponse:
    logger.debug("Perform %s %s on %s:%d", method, url, host, port)
    try:
        async with aiohttp.ClientSession(
            f"https://{host}:{port}", connector=TCPConnector(ssl=SSL_CONTEXT)
        ) as session, session.request(method, url, json=body, headers=create_headers(host, token)) as response_http:
            response_status = response_http.status
            response_body = await response_http.text()
            logger.debug("Response status=%d body=%s", response_status, response_body)

            if response_status not in do_not_raise_exception_on:
                raise Exception(f"Call failed (status={response_status} body={response_body}")

            return HttpResponse(response_status, response_body)
    except aiohttp.client_exceptions.ClientConnectorError as e:
        raise ConnectionFailed(f"Failed to connect to host: {host}") from e



from typing import Optional

from .frames import parse_stomp_frame
from .frames_models import StompFrame
from .frames_server_models import ConnectedFrame, ErrorFrame, MessageFrame, ReceiptFrame


def parse_server_frame(content: str) -> Optional[StompFrame]:
    stomp_frame = parse_stomp_frame(content)
    if stomp_frame is None:
        return None

    match stomp_frame.message:
        case "CONNECTED":
            return ConnectedFrame(stomp_frame)
        case "MESSAGE":
            return MessageFrame(stomp_frame)
        case "RECEIPT":
            return ReceiptFrame(stomp_frame)
        case "ERROR":
            return ErrorFrame(stomp_frame)
        case _:
            raise Exception("Unknown message type", stomp_frame.message)



import asyncio
from typing import Optional


def to_int(value: Optional[str]) -> Optional[int]:
    return int(value) if value else None


def current_task_is_running() -> bool:
    task = asyncio.current_task()
    if task is None:
        return False
    return not task.cancelling() and not task.cancelled()



from typing import List, Optional

from .interior_unit_models import InteriorUnit


def compute_interior_unit_diff_description(old: Optional[InteriorUnit], new: Optional[InteriorUnit]) -> str:
    if old == new:
        return "No change"

    if old is None:
        return f"Created {new}"

    if new is None:
        return f"Deleted {old.id}"

    changes: List[str] = []
    if old.id != new.id:
        changes.append(f"id: {old.id}=>{new.id}")
    if old.name != new.name:
        changes.append(f"name: {old.name}=>{new.name}")
    if old.power != new.power:
        changes.append(f"power: {old.power}=>{new.power}")
    if old.mode != new.mode:
        changes.append(f"mode: {old.mode}=>{new.mode}")
    if old.requested_temperature != new.requested_temperature:
        changes.append(f"requested_temperature: {old.requested_temperature}=>{new.requested_temperature}")
    if old.humidity != new.humidity:
        changes.append(f"humidity: {old.humidity}=>{new.humidity}")
    if old.fan_speed != new.fan_speed:
        changes.append(f"fan_speed: {old.fan_speed}=>{new.fan_speed}")
    if old.fan_swing != new.fan_swing:
        changes.append(f"fan_swing: {old.fan_swing}=>{new.fan_swing}")
    if old.room_temperature != new.room_temperature:
        changes.append(f"room_temperature: {old.room_temperature}=>{new.room_temperature}")
    if old.updated_at != new.updated_at:
        changes.append(f"updated_at: {old.updated_at}=>{new.updated_at}")
    if old.online != new.online:
        changes.append(f"online: {old.online}=>{new.online}")
    if old.online_updated_at != new.online_updated_at:
        changes.append(f"online_updated_at: {old.online_updated_at}=>{new.online_updated_at}")

    return "Changed " + ", ".join(changes)



import logging
from typing import List, Optional

from aircloudy.contants import DEFAULT_REST_API_HOST, FanSpeed, FanSwing, OperatingMode, Power

from ..errors import TooManyRequestsException
from ..interior_unit_models import InteriorUnit
from .http_client import perform_request
from .rac_models import CommandResponse, CommandStatus, PowerAllResponse, _GeneralControlCommand, _InteriorUnitRest

logger = logging.getLogger(__name__)


async def get_interior_units(
    token: str, family_id: int, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> List[InteriorUnit]:
    logger.debug("Get interior units")
    response = await perform_request(
        "GET", f"/rac/ownership/groups/{family_id}/idu-list", token=token, host=host, port=port
    )

    if response.status != 200:
        raise Exception(f"Call failed (status={response.status} body={response.body}")

    return [_InteriorUnitRest(d).to_internal_representation() for d in response.body_as_json]


async def get_command_status(
    token: str, commands: List[CommandResponse], host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> List[CommandStatus]:
    logger.debug("Get command status")

    response = await perform_request(
        "POST", "/rac/status/command", [c.__dict__ for c in commands], token=token, host=host, port=port
    )

    return [CommandStatus(s) for s in response.body_as_json]


async def configure_interior_unit(
    token: str,
    family_id: int,
    interior_unit: InteriorUnit,
    power: Optional[Power] = None,
    mode: Optional[OperatingMode] = None,
    requested_temperature: Optional[float] = None,
    humidity: Optional[int] = None,
    fan_speed: Optional[FanSpeed] = None,
    fan_swing: Optional[FanSwing] = None,
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> CommandResponse:
    """Send command to change interior unit (like a remote control)

    :raises:
        TooManyRequestsException: If previous command is still in progress
    """
    command = _GeneralControlCommand(
        interior_unit.copy(
            power=power,
            mode=mode,
            requested_temperature=requested_temperature,
            humidity=humidity,
            fan_speed=fan_speed,
            fan_swing=fan_swing,
        )
    )

    logger.debug("Configure interior unit familiy_id=%s : %s", family_id, command)
    response = await perform_request(
        "PUT",
        f"/rac/basic-idu-control/general-control-command/{command.id}?familyId={family_id}",
        command.__dict__,
        do_not_raise_exception_on=(200, 429),
        token=token,
        host=host,
        port=port,
    )

    if response.status == 429:
        raise TooManyRequestsException(response.body)

    return CommandResponse(response.body_as_json)


async def request_refresh_interior_unit_state(
    token: str, rac_id: int, family_id: int, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> None:
    logger.debug("Request refresh interior unit state for rac id=%s, family_id=%s", rac_id, family_id)
    await perform_request("PUT", f"/rac/status/{rac_id}?familyId={family_id}", token=token, host=host, port=port)


async def set_power(token: str, rac_id: str, power: Power, host: str = DEFAULT_REST_API_HOST, port: int = 443) -> None:
    logger.debug("Set power rac_id=%s, power=%s", rac_id, power)
    await perform_request(
        "PUT",
        f"/rac/basic-idu-control/switch-on-off/{rac_id}",
        {
            "power": power,
        },
        token=token,
        host=host,
        port=port,
    )


async def set_power_all(
    token: str,
    family_id: str,
    power: Power,
    interior_units: List[InteriorUnit],
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> PowerAllResponse:
    match power:
        case "ON":
            url = f"/rac/manage-idu/groups/{family_id}/idu/start"
        case "OFF":
            url = f"/rac/manage-idu/groups/{family_id}/idu/stop"
        case _:
            raise Exception(f"Unknown power value {power}")

    units = [_GeneralControlCommand(iu.copy(power=power)).__dict__ for iu in interior_units]

    logger.debug("Set power all power=%s for %s", power, units)
    response = await perform_request(
        "PUT", url, units, do_not_raise_exception_on=(200, 207), token=token, host=host, port=port
    )

    return PowerAllResponse(response.body_as_json)



from typing import Optional


class StompFrame:
    message: str
    headers: dict[str, str]
    body: Optional[dict] = None

    def __init__(self, message: str, headers: dict[str, str], body: Optional[dict] = None) -> None:
        self.message = message
        self.headers = headers
        self.body = body

    def __repr__(self) -> str:
        return f"WebsocketMessage(message={self.message}, headers={self.headers}, body={self.body})"

    def get_frame(self) -> str:
        websocket_message = f"{self.message}\n"
        for name, value in self.headers.items():
            websocket_message += f"{name}:{value}\n"
        websocket_message += "\n"

        if self.body:
            websocket_message += f"{self.body}\n"

        return websocket_message + "\0"



import asyncio
import logging
from asyncio import Task
from typing import Callable, Optional

from aircloudy.api.rac import get_command_status
from aircloudy.api.rac_models import CommandResponse, CommandStatus
from aircloudy.contants import DEFAULT_REST_API_HOST, DEFAULT_WAIT_DONE, CommandState
from aircloudy.utils import current_task_is_running

logger = logging.getLogger(__name__)


class CommandManager:
    _token_supplier: Callable[[], str]
    _update_interval: int
    _api_host: str
    _port: int

    _commands: dict[str, CommandResponse]
    _commands_status: dict[str, CommandStatus]
    _events: dict[str, asyncio.Event]
    _task_fetch_command_status: Optional[Task]

    def __init__(
        self,
        token_supplier: Callable[[], str],
        update_interval: int = 2,
        host: str = DEFAULT_REST_API_HOST,
        port: int = 443,
    ) -> None:
        self._token_supplier = token_supplier
        self._update_interval = update_interval
        self._api_host = host
        self._port = port

        self._commands = {}
        self._commands_status = {}
        self._events = {}
        self._task_fetch_command_status = None

    def add_command_watch(self, command: CommandResponse) -> None:
        self._commands[command.commandId] = command
        if self._task_fetch_command_status is None:
            self._task_fetch_command_status = asyncio.create_task(self._fetch_command_status_loop())

    def remove_command_watch(self, command_id: str) -> None:
        del self._commands[command_id]
        if self._task_fetch_command_status is not None and len(self._commands) == 0:
            self._task_fetch_command_status.cancel()

    def clear(self) -> None:
        self._commands = {}
        if self._task_fetch_command_status is not None:
            self._task_fetch_command_status.cancel()

    async def wait_ack(self, command: CommandResponse, wait: int = DEFAULT_WAIT_DONE) -> CommandState:
        self.add_command_watch(command)
        event = asyncio.Event()
        self._events[command.commandId] = event
        try:
            await asyncio.wait_for(event.wait(), wait)
            del self._events[command.commandId]
            self.remove_command_watch(command.commandId)
        except asyncio.TimeoutError:
            pass

        status = self._commands_status[command.commandId]
        return status.status

    async def _fetch_command_status_loop(self) -> None:
        while current_task_is_running():
            try:
                logger.debug("Fetch tasks status")
                self._commands_status = {
                    s.commandId: s
                    for s in await get_command_status(
                        self._token_supplier(), list(self._commands.values()), self._api_host, self._port
                    )
                }
                for command_id, event in self._events.items():
                    if not event.is_set():
                        status = self._commands_status[command_id]
                        if status is not None and status.status == "DONE":
                            event.set()

                await asyncio.sleep(self._update_interval)
            except Exception as e:
                logger.error("Unexpected error while fetching status : %s", e)
                raise e



from .aircloud import HitachiAirCloud
from .contants import FanSpeed, FanSwing, OperatingMode, Power, ScheduleType
from .errors import (
    AuthenticationFailedException,
    CommandFailedException,
    ConnectionFailed,
    IllegalStateException,
    InteriorUnitNotFoundException,
    TooManyRequestsException,
)
from .interior_unit import compute_interior_unit_diff_description
from .interior_unit_models import InteriorUnit



from .frames import WebsocketMessageParseException, parse_stomp_frame
from .frames_models import StompFrame
from .frames_server import parse_server_frame
from .frames_server_models import ConnectedFrame, ErrorFrame, MessageFrame, ReceiptFrame



import json
from io import StringIO
from typing import Optional

from .frames_models import StompFrame


class WebsocketMessageParseException(Exception):
    error: str
    message: str
    line: str

    def __init__(self, error: str, message: str, line: str) -> None:
        Exception.__init__(self)
        self.error = error
        self.message = message
        self.line = line


def parse_stomp_frame(content: str) -> Optional[StompFrame]:
    io = StringIO(content)

    first_line = io.readline()
    if first_line == "":
        raise WebsocketMessageParseException("Message is empty", content, first_line)
    if first_line == "\n":
        return None

    message = first_line.strip("\n")
    headers = {}
    while True:
        line = io.readline()
        if line in ("", "\n"):
            break

        header = line.strip("\n").split(":", 1)
        headers[header[0]] = header[1]

    body_raw = io.read().strip("\0")

    if body_raw == "":
        return StompFrame(message, headers, None)

    return StompFrame(message, headers, json.loads(body_raw))



from dataclasses import dataclass

from aircloudy.contants import FanSpeed, FanSwing, OperatingMode, Power, ScheduleType
from aircloudy.interior_unit_models import InteriorUnit


class IduFrostWashStatus:
    active: bool
    priority: int
    astUpdatedA: int
    subCategory = None
    errorCode = None


class SpecialOperationStatus:
    active: bool
    priority: int
    lastUpdatedAt: int
    subCategory = None
    errorCode = None


class ErrorStatus:
    active: bool
    priority: int
    lastUpdatedAt: int
    subCategory: str
    errorCode = None


class HolidayModeStatus:
    active: bool
    priority: int
    lastUpdatedAt: int
    subCategory = None
    errorCode = None


@dataclass
class InteriorUnitNotification:
    serialNumber: str
    iduFrostWashStatus: IduFrostWashStatus
    model: str
    id: int
    vendorThingId: str
    name: str
    roomTemperature: float
    mode: OperatingMode
    iduTemperature: float
    humidity: int
    power: Power
    relativeTemperature: float
    fanSpeed: FanSpeed
    fanSwing: FanSwing
    updatedAt: int
    specialOperationStatus: SpecialOperationStatus
    errorStatus: ErrorStatus
    lastOnlineUpdatedAt: int
    scheduletype: ScheduleType
    modelTypeId: int
    cloudId: str
    opt4: int
    holidayModeStatus: HolidayModeStatus
    online: bool
    SysType: int

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)

    def to_internal_representation(self) -> InteriorUnit:
        return InteriorUnit(
            self.id,
            self.name,
            self.power,
            self.mode,
            self.iduTemperature,
            self.humidity,
            self.fanSpeed,
            self.fanSwing,
            self.roomTemperature,
            self.updatedAt,
            self.online,
            self.lastOnlineUpdatedAt,
        )
